package de.hdm_stuttgart.mi.codeformat;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.lang.reflect.Method;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.apache.log4j.Logger;
import org.w3c.dom.DOMConfiguration;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;
import org.w3c.dom.bootstrap.DOMImplementationRegistry;
import org.w3c.dom.ls.DOMImplementationLS;
import org.w3c.dom.ls.LSOutput;
import org.w3c.dom.ls.LSParser;
import org.w3c.dom.ls.LSSerializer;


/** 
 * @author $Author: goik $
 * @version $Rev: 289 $ $Date: 2007-11-19 12:05:47 +0100 (Mon, 19 Nov 2007) $
 *
 * $URL: https://version/my.java $.
 *
 */
public class Docbook2docbook {
	
	private static final Logger logger = Logger.getLogger(Docbook2docbook.class);
  private final static String srcPrefixDir = "Ref/api",
                          docbookNamespace = "http://docbook.org/ns/docbook";

  private final XPathFactory xpf = XPathFactory.newInstance();
  
  /** Searching a document of type docbook for programlistings with java classes. If the class can be retrieved
   * a url reference is being constructed.
   * @param args
   * @throws ClassCastException
   * @throws FileNotFoundException
   * @throws ClassNotFoundException
   * @throws InstantiationException
   * @throws IllegalAccessException
   * @throws XPathExpressionException
   */
  public static void main(String[] args) throws 
      ClassCastException, FileNotFoundException, ClassNotFoundException,
      InstantiationException, IllegalAccessException, XPathExpressionException {
    if (1 != args.length) {
      logger.fatal("Usage: Docbook2docbook <xmlfileBasename>");
    } else {
      Docbook2docbook d2d = new Docbook2docbook(args[0]);
      d2d.process();
      logger.info("Successfully finished conversion");
    }
  }
  private 
  final String docbookBasename;
  /**
   * @param basename the XML document's basename e.g. for a XML instance "somdirectory/mydoc.xml"
   * this will be "somdirectory/mydoc"
   */
  public Docbook2docbook (final String basename){
    this.docbookBasename = basename;
  }
  private void process() throws ClassCastException, ClassNotFoundException,
   InstantiationException, IllegalAccessException,
   XPathExpressionException {
    final DOMImplementationRegistry registry =
      DOMImplementationRegistry.newInstance();

    DOMImplementationLS impl = (DOMImplementationLS)registry.getDOMImplementation("LS");

    LSParser db = impl.createLSParser(DOMImplementationLS.MODE_SYNCHRONOUS, null);
    final DOMConfiguration config = db.getDomConfig();
    config.setParameter("error-handler", new DomErrorHandler(System.err));
    final String xmlInfileName = docbookBasename + ".xml";
    final Document dbInput = db.parseURI(xmlInfileName);
    addClassReferences(dbInput);
    
    final LSSerializer serial = impl.createLSSerializer();
    serial.getDomConfig().setParameter("format-pretty-print", false);
    final LSOutput lsOut = impl.createLSOutput();
    PrintWriter out;
    try {
      out = new PrintWriter(docbookBasename + ".convert.xml");
      lsOut.setCharacterStream(out);
      serial.write(dbInput, lsOut);
      out.flush();
    } catch (FileNotFoundException e) {
      logger.fatal("Unable to parse xml file " + xmlInfileName);
    }
  }
  
  private void addClassReferences(final Document top) throws XPathExpressionException{
    
    final XPath searchBlockObjects = xpf.newXPath();
    searchBlockObjects.setNamespaceContext(new MyNamespaceContexts());

    final NodeList classnames = 
        (NodeList) searchBlockObjects.evaluate(
          "//db:classname[not(contains(@annotations, 'nojavadoc'))]",
            top, XPathConstants.NODESET);
      for (int i = 0; i < classnames.getLength(); i++){
          processClassName((Element) classnames.item(i));
      }

      final NodeList listings = 
        (NodeList) searchBlockObjects.evaluate(
          "//db:programlisting[not(contains(@annotations, 'nojavadoc'))]",
            top, XPathConstants.NODESET);
      for (int i = 0; i < listings.getLength(); i++){
        processListing((Element) listings.item(i));
      }      
        // Not yet ready!
//        final NodeList methodNames = 
//            (NodeList) searchProgramlisting.evaluate(
//              "//db:methodname",
//                top, XPathConstants.NODESET);
//          for (int i = 0; i < methodNames.getLength(); i++){
//              processMethodName((Element) methodNames.item(i));
//          }
  }
  
  private void processClassName(final Element classnameElement) {
    final Node firstClassnameNode =  classnameElement.getFirstChild();
    if (firstClassnameNode instanceof Text) {
      final Text firstClassnameTextNode = (Text) firstClassnameNode;
      final String fullyQualifiedClassname = firstClassnameTextNode.getData().trim();
      
      final Class<?> cl = testClassAvailability(fullyQualifiedClassname);
      if (null == cl) {
    	  logger.warn("Unable to find prefix for '" + fullyQualifiedClassname);
      } else {
    	  final String urlPrefix, implementationVendor;
    	  if (fullyQualifiedClassname.startsWith("javax.")) {
    		  urlPrefix="http://docs.oracle.com/javase/7/docs/api";
    		  implementationVendor = "oracle corporation (fixed by javax.*)";
    	  } else {
    		  implementationVendor = cl.getPackage().getImplementationVendor();
    		  if (null == implementationVendor) {
    			  urlPrefix = srcPrefixDir;
    		  } else {
    			  switch(implementationVendor.toLowerCase()) {
    			  case "oracle corporation":
    				  urlPrefix="http://docs.oracle.com/javase/7/docs/api";
    				  break;
    			  case "jdom.org":
    				  urlPrefix="http://www.jdom.org/docs/apidocs";
    				  break;
    			  case "hibernate.org":
    				  urlPrefix = "http://docs.jboss.org/hibernate/orm/4.1/javadocs";
    				  break;
    			  default:
    				  urlPrefix="unknown vendor '" + implementationVendor + "'";
    				  logger.warn("No vendor found for " + fullyQualifiedClassname);
    				  break;
    			  }
    		  }
    		  final Element link = createLink(classnameElement, cl, cl.getSimpleName());
    		  link.setAttributeNS("http://www.w3.org/1999/xlink", "xlink:href",
    				  urlPrefix + "/" + package2path(cl.getCanonicalName()) + ".html");
    		  classnameElement.replaceChild(link, firstClassnameTextNode);
    	  }
    	  logger.info("Mapping '" + fullyQualifiedClassname + "' to base URL " +
    			  urlPrefix + ", vendor = " + implementationVendor);
      }
    }
  }
  
  private void processMethodName(final Element methodNameElement) {
    final Node firstMethodNameNode =  methodNameElement.getFirstChild();
    if (firstMethodNameNode instanceof Text) {
      final Text firstMethodNameTextNode = (Text) firstMethodNameNode;
      final String fullyQualifiedMethodName = firstMethodNameTextNode.getData().trim();
//      final Class<?> cl = testClassAvailability(fullyQualifiedClassname);
//      if (null != cl) {
//        final String urlPrefix;
//        final String implementationVendor = cl.getPackage().getImplementationVendor();
//        if (null != implementationVendor && implementationVendor.equals("Oracle Corporation")){
//          urlPrefix="http://docs.oracle.com/javase/7/docs/api";
//        } else {
//          urlPrefix = srcPrefixDir;
//        }
//        final Element link = createLink(classnameElement, cl, cl.getSimpleName());
//        link.setAttributeNS("http://www.w3.org/1999/xlink", "xlink:href",
//            urlPrefix + "/" + package2path(cl.getCanonicalName()) + ".html");
//        classnameElement.replaceChild(link, firstClassnameTextNode);
//      }
    }
  }

  private void processListing(final Element plisting) throws XPathExpressionException{
    final Text[] textNodes = findTextNodes(plisting);
    
    final Pattern searchPackage = Pattern.compile(
        "(?s).*package[ \\t]+" + 
        "([\\w]|\\w[\\w|.]*[\\w])"+
        "[ \t]*;.*");

    String packageName = null;
    for (final Text tNode : textNodes){
      Matcher matchPackage = searchPackage.matcher(tNode.getData());
      if (matchPackage.find()){
        packageName = matchPackage.group(1);
        break;
      }
    }
    if (null != packageName){
      final Pattern searchClass = Pattern.compile(
          "(?s)(.*[ \\t]+class[ \\t]+)" +
          "(\\w+)" +
          "(.*)");
      // May use the following regexp excluding non-public classes with group(4):
      //      "(?s)(.*public)([ \\t]+\\w+)*[ \\t]+(class[ \\t]+)" +
      //      "(\\w+)" +
      //      "(.*)");
      for (final Text tNode : textNodes){
        Matcher matchClass = searchClass.matcher(tNode.getData());
        if (matchClass.find()){
          final String className = matchClass.group(2);

          final Class<?> cl = testClassAvailability(packageName + "." + className);
          if (null != cl){
            final Text beforeText = createTextNode(plisting, matchClass.group(1)),
            afterText = createTextNode(plisting, matchClass.group(3));
            plisting.replaceChild(afterText, tNode);
            final Element link = createLink(plisting, cl, className);
            link.setAttributeNS("http://www.w3.org/1999/xlink", "xlink:href",
                srcPrefixDir + "/" + package2path(cl.getCanonicalName()) + ".html");
            plisting.insertBefore(link, afterText);
            plisting.insertBefore(beforeText, link);
          }
          break;
        }
      }
    }
  }
  private Element createLink(final Element root, final Class<?> cl, final String className) {
    final Element link = createElementNode(
        root,
        "uri", docbookNamespace,
        className);
    link.setAttributeNS("http://www.w3.org/1999/xlink", "xlink:href",
        srcPrefixDir + "/" + package2path(cl.getCanonicalName()) + ".html");
    return link;
  }
  private Class<?> testClassAvailability(final String fullyQualifiedClassName){
      try {
        return Class.forName(fullyQualifiedClassName);
      } catch (ClassNotFoundException e) {
        if (fullyQualifiedClassName.contains(".")) {
          logger.warn("Warning: Unable to access class " + fullyQualifiedClassName);
        }
        return null;
      }
  }
  //Helper functions
  
  /**
   * @param current
   * @return
   */
  public String findExtraClasspath(final Element current) {
    
    logger.info("line:" + current.getUserData("lineNumber"));
//    final String annotations = current.getAttributeNS(docbookNamespace, "annotations");
    final String annotations = current.getAttribute("annotations");
    if (null != annotations) {
      final String ret = annotations.trim();
      if (!ret.equals("")) {
        return ret;
      } 
    }
    final Node parentNode = current.getParentNode();
    
    if (parentNode instanceof Document) {
      logger.error("Document root being reached while scanning for classpath related annotation");
      System.exit(1);
    }
    final Element parent = (Element) parentNode;
    return findExtraClasspath(parent);
  }
  
  /**
   * @param path
   */
  public static void addURL(String path)  {
    File filePath = new File(path);
    try {
      addURL(filePath.toURI().toURL());
    } catch (MalformedURLException e) {
      e.printStackTrace();
      logger.fatal("Error, could not create URL from '" + path + "'");
      System.exit(1);
    }
  }

  /**
   * @param u
   */
  public static void addURL(URL u) {

    URLClassLoader sysloader = (URLClassLoader) ClassLoader.getSystemClassLoader();
    Class<URLClassLoader> sysclass = URLClassLoader.class;
    final Class[] parameters = new Class[]{URL.class};

    try {
      Method method = sysclass.getDeclaredMethod("addURL", parameters);
      method.setAccessible(true);
      method.invoke(sysloader, new Object[]{u});
    } catch (Throwable t) {
      t.printStackTrace();
      logger.error("Error, could not add URL to system classloader");
      System.exit(1);
    }
  }
  
  private final String package2path(final String packageName){
    String searchDot = "\\.";
    Pattern pattern = Pattern.compile(searchDot);
    Matcher matcher = pattern.matcher(packageName);
    return matcher.replaceAll("/");
  }
  private Text createTextNode(final Node node, final String data){
    if (node instanceof Document){
      return ((Document) node).createTextNode(data);
    } else{
      return node.getOwnerDocument().createTextNode(data);
    }
  }
  private Element createElementNode(final Node node,
      final String type, final String ns, final String data){
    final Element ret;
    if (node instanceof Document){
      ret = ((Document) node).createElementNS(ns, type);
    } else{
      ret = node.getOwnerDocument().createElementNS(ns, type);
    }
    final Text t = createTextNode(node, data);
    ret.appendChild(t);
    return ret;
  }
  private Text[] findTextNodes(final Element parent) throws XPathExpressionException{
    final XPath searchTextNodes = xpf.newXPath();
    final NodeList textNodes = 
      (NodeList) searchTextNodes.evaluate(
        "text()",
          parent, XPathConstants.NODESET);
    final Text [] t = new Text[textNodes.getLength()];
    for (int i = 0; i < textNodes.getLength(); i++){
      t[i] = (Text) textNodes.item(i);
    }
    return t;
  }
}
